package cn.iocoder.yudao.module.system.service.sms;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.StrUtil;
import cn.iocoder.yudao.framework.apollo.core.constants.Constants;
import cn.iocoder.yudao.framework.common.util.json.core.KeyValue;
import cn.iocoder.yudao.framework.common.util.spring.enums.CommonStatusEnum;
import cn.iocoder.yudao.framework.common.util.spring.enums.UserTypeEnum;
import cn.iocoder.yudao.framework.redis.helper.RedisHelper;
import cn.iocoder.yudao.framework.sms.core.client.SmsClient;
import cn.iocoder.yudao.framework.sms.core.client.SmsClientFactory;
import cn.iocoder.yudao.framework.sms.core.client.SmsCommonResult;
import cn.iocoder.yudao.framework.sms.core.client.dto.SmsReceiveRespDTO;
import cn.iocoder.yudao.framework.sms.core.client.dto.SmsSendRespDTO;
import cn.iocoder.yudao.framework.sms.core.client.dto.SmsTemplateDTO;
import cn.iocoder.yudao.framework.web.config.BusinessProperties;
import cn.iocoder.yudao.module.member.api.user.MemberUserApi;
import cn.iocoder.yudao.module.member.api.user.dto.UserRespDTO;
import cn.iocoder.yudao.module.system.convert.sms.SmsTemplateConvert;
import cn.iocoder.yudao.module.system.dal.dataobject.sms.SmsChannelDO;
import cn.iocoder.yudao.module.system.dal.dataobject.sms.SmsLogDO;
import cn.iocoder.yudao.module.system.dal.dataobject.sms.SmsNodeDO;
import cn.iocoder.yudao.module.system.dal.dataobject.sms.SmsTemplateDO;
import cn.iocoder.yudao.module.system.dal.dataobject.user.AdminUserDO;
import cn.iocoder.yudao.module.system.enums.sms.SmsCountryCodeEnum;
import cn.iocoder.yudao.module.system.enums.sms.SystemStatusEnum;
import cn.iocoder.yudao.module.system.mq.message.sms.SmsSendMessage;
import cn.iocoder.yudao.module.system.mq.message.sms.SmsSendMessageV2;
import cn.iocoder.yudao.module.system.mq.producer.sms.SmsProducer;
import cn.iocoder.yudao.module.system.service.user.AdminUserService;
import com.alibaba.fastjson.JSON;
import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;

import javax.annotation.Resource;
import java.util.*;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.system.enums.ErrorCodeConstants.*;

/**
 * 短信发送 Service 发送的实现
 *
 * @author 捷道源码
 */
@Slf4j
@Service
public class SmsSendServiceImpl implements SmsSendService {

    @Resource
    private AdminUserService adminUserService;
    @Resource
    private MemberUserApi memberUserApi;

    @Resource
    private SmsTemplateService smsTemplateService;

    @Resource
    private SmsChannelService smsChannelService;

    @Resource
    private SmsLogService smsLogService;

    @Resource
    private SmsClientFactory smsClientFactory;

    @Resource
    private SmsProducer smsProducer;
    @Resource
    private BusinessProperties businessProperties;

    @Resource
    private SmsNodeService smsNodeService;

    @Resource
    private RedisHelper redisHelper;

    @Override
    public Long sendSingleSmsToAdmin(String mobile, Long userId, String templateCode, Map<String, Object> templateParams, String areaCode) {
        // 如果 mobile 为空，则加载用户编号对应的手机号
        if (StrUtil.isEmpty(mobile)) {
            AdminUserDO user = adminUserService.getUser(userId);
            if (user != null) {
                mobile = user.getMobile();
                areaCode = "86"; //TODO 待管理端用户信息存入国家区号
            }
        }
        // 执行发送
        return this.sendSingleSms(areaCode + mobile, userId, UserTypeEnum.ADMIN.getValue(), templateCode, templateParams);
    }

    @Override
    public Long sendSingleSmsToAdminV2(String mobile, Long userId, String nodeValue, String areaCode,
                                       Integer isOrders, Integer isTransport, Integer transportId, Integer messageType, Map<String, Object> templateParams, String receive) {
        // 如果 mobile 为空，则加载用户编号对应的手机号
        if (StrUtil.isEmpty(mobile)) {
            AdminUserDO user = adminUserService.getUser(userId);
            if (user != null) {
                mobile = user.getMobile();
                areaCode = "86"; //TODO 待管理端用户信息存入国家区号
            }
        }
        // 执行发送
        return this.sendSingleSmsV2(areaCode + mobile, userId, UserTypeEnum.ADMIN.getValue(), nodeValue, areaCode, isOrders, isTransport, transportId, messageType, templateParams, null, 1, null, receive);
    }

    @Override
    public Long sendSingleSmsToMember(String mobile, Long userId, String templateCode, Map<String, Object> templateParams, String areaCode) {
        // 如果 mobile 为空，则加载用户编号对应的手机号
        if (StrUtil.isEmpty(mobile)) {
            UserRespDTO user = memberUserApi.getUser(userId);
            if (user != null) {
                mobile = user.getMobile();
                areaCode = user.getAreaCode(); //TODO 待管理端用户信息存入国家区号
            }
        }
        // 执行发送
        return this.sendSingleSms(areaCode + mobile, userId, UserTypeEnum.MEMBER.getValue(), templateCode, templateParams);
    }

    @Override
    public Long sendSingleSmsToMemberV2(String mobile, Long userId, String nodeValue, String areaCode,
                                        Integer isOrders, Integer isTransport, Integer transportId, Integer messageType, Map<String, Object> templateParams, String receive) {
        // 如果 mobile 为空，则加载用户编号对应的手机号
        if (StrUtil.isEmpty(mobile)) {
            UserRespDTO user = memberUserApi.getUser(userId);
            if (user != null) {
                mobile = user.getMobile();
                areaCode = user.getAreaCode(); //TODO 待管理端用户信息存入国家区号
            }
        }
        // 执行发送
        return this.sendSingleSmsV2(areaCode + mobile, userId, UserTypeEnum.MEMBER.getValue(), nodeValue, areaCode, isOrders, isTransport, transportId, messageType, templateParams, null, 1, null, receive);
    }

    @Override
    public Long sendSingleSms(String mobile, Long userId, Integer userType,
                              String templateCode, Map<String, Object> templateParams) {
        // 校验短信模板是否合法
        SmsTemplateDO template = this.checkSmsTemplateValid(templateCode);
        // 校验手机号码是否存在
        mobile = this.checkMobile(mobile);
        // 构建有序的模板参数。为什么放在这个位置，是提前保证模板参数的正确性，而不是到了插入发送日志
        List<KeyValue<String, Object>> newTemplateParams = this.buildTemplateParams(template, templateParams);

        // 创建发送日志。如果模板被禁用，则不发送短信，只记录日志
        Boolean isSend = CommonStatusEnum.ENABLE.getStatus().equals(template.getStatus());
        String content;
        String apiTemplateId;
        if (mobile.startsWith("0086") || mobile.startsWith("86") || mobile.startsWith("+86") || mobile.startsWith("+0086")) {
            content = smsTemplateService.formatSmsTemplateContent(template.getContent(), templateParams);
            apiTemplateId = template.getApiTemplateId();
        } else {
            //国外
            content = smsTemplateService.formatSmsTemplateContent(template.getContentEn(), templateParams);
            apiTemplateId = template.getApiTemplateIdEn();
        }

        Long sendLogId = smsLogService.createSmsLog(mobile, userId, userType, isSend, template, content, templateParams, apiTemplateId, null);

        Long channelId = template.getChannelId();
        SmsChannelDO smsChannel = smsChannelService.getSmsChannel(channelId);
        if (smsChannel == null) {
            throw exception(SMS_CHANNEL_NOT_EXISTS);
        }

        // 发送 MQ 消息，异步执行发送短信 (补充debug模式启用时，不执行真实发送任务. 补充渠道关闭不执行真实发送任务)
        if (isSend && (Objects.isNull(businessProperties) || !businessProperties.isDebug()) && smsChannel.getStatus() == 0) {
            smsProducer.sendSmsSendMessage(sendLogId, mobile, channelId,
                    apiTemplateId, newTemplateParams);
        }
        return sendLogId;
    }


    @Override
    public void resendSingleSmsBySmsLogIds(Collection<Long> smsLogIds) {
        if (CollectionUtil.isEmpty(smsLogIds)) {
            return;
        }
        List<SmsLogDO> smsLogDOList = smsLogService.listByIds(smsLogIds);
        if (CollectionUtil.isEmpty(smsLogDOList)) {
            return;
        }
        List<SmsLogDO> resendSmsLogList = new ArrayList<>();
        for (SmsLogDO smsLogDO : smsLogDOList) {
            String mobile = smsLogDO.getMobile();
            // 校验短信模板是否合法
            SmsTemplateDO template = this.checkSmsTemplateValid(smsLogDO.getTemplateCode());
            // 创建发送日志。如果模板被禁用，则不发送短信，只记录日志
            Boolean isSend = CommonStatusEnum.ENABLE.getStatus().equals(template.getStatus());
            String apiTemplateId = smsLogDO.getApiTemplateId();
            Long sendLogId = smsLogService.createSmsLog(mobile, smsLogDO.getUserId(), smsLogDO.getUserType(),
                    isSend, template, smsLogDO.getTemplateContent(), smsLogDO.getTemplateParams(), apiTemplateId,
                    Objects.isNull(smsLogDO.getSmsLogId()) ? smsLogDO.getId() : smsLogDO.getSmsLogId());
            // 发送 MQ 消息，异步执行发送短信 (补充debug模式启用时，不执行真实发送任务)
            if (isSend && (Objects.isNull(businessProperties) || !businessProperties.isDebug())) {
                // 构建有序的模板参数。提前保证模板参数的正确性，而不是到了插入发送日志
                List<KeyValue<String, Object>> newTemplateParams = this.buildTemplateParams(template, smsLogDO.getTemplateParams());
                smsProducer.sendSmsSendMessage(sendLogId, mobile, template.getChannelId(),
                        apiTemplateId, newTemplateParams);
            }
            if (Objects.isNull(smsLogDO.getSmsLogId())) {
                smsLogDO.setNum(smsLogDO.getNum() + 1);
                resendSmsLogList.add(smsLogDO);
            } else {
                // 获取原始重发短信日志
                SmsLogDO orgSmsLogDO = smsLogService.getById(smsLogDO.getSmsLogId());
                orgSmsLogDO.setNum(orgSmsLogDO.getNum() + 1);
                resendSmsLogList.add(orgSmsLogDO);
            }
        }
        if (CollectionUtil.isNotEmpty(resendSmsLogList)) {
            smsLogService.updateBatchSmsLog(resendSmsLogList);
        }
    }

    @VisibleForTesting
    public SmsTemplateDO checkSmsTemplateValid(String templateCode) {
        // 获得短信模板。考虑到效率，从缓存中获取
        SmsTemplateDO template = smsTemplateService.getSmsTemplateByCodeFromCache(templateCode);
        // 短信模板不存在
        if (template == null) {
            throw exception(SMS_SEND_TEMPLATE_NOT_EXISTS);
        }
        return template;
    }

    /**
     * 将参数模板，处理成有序的 KeyValue 数组
     * <p>
     * 原因是，部分短信平台并不是使用 key 作为参数，而是数组下标，例如说腾讯云 <a href="https://cloud.tencent.com/document/product/382/39023">...</a>
     *
     * @param template       短信模板
     * @param templateParams 原始参数
     * @return 处理后的参数
     */
    @VisibleForTesting
    public List<KeyValue<String, Object>> buildTemplateParams(SmsTemplateDO template, Map<String, Object> templateParams) {
        return template.getParams().stream().map(key -> {
            Object value = templateParams.get(key);
            if (value == null) {
                throw exception(SMS_SEND_MOBILE_TEMPLATE_PARAM_MISS, key);
            }
            return new KeyValue<>(key, value);
        }).collect(Collectors.toList());
    }

    @VisibleForTesting
    public String checkMobile(String mobile) {
        if (StrUtil.isEmpty(mobile)) {
            throw exception(SMS_SEND_MOBILE_NOT_EXISTS);
        }
        return mobile;
    }

    @Override
    public void doSendSms(SmsSendMessage message) {
        // 获得渠道对应的 SmsClient 客户端
        SmsClient smsClient = smsClientFactory.getSmsClient(message.getChannelId());
        Assert.notNull(smsClient, String.format("短信客户端(%d) 不存在", message.getChannelId()));
        // 发送短信
        SmsCommonResult<SmsSendRespDTO> sendResult = smsClient.sendSms(message.getLogId(), message.getMobile(),
                message.getApiTemplateId(), message.getTemplateParams());
        smsLogService.updateSmsSendResult(message.getLogId(), sendResult.getCode(), sendResult.getMsg(),
                sendResult.getApiCode(), sendResult.getApiMsg(), sendResult.getApiRequestId(),
                sendResult.getData() != null ? sendResult.getData().getSerialNo() : null);
    }

    @Override
    public void doSendSmsV2(SmsSendMessageV2 message) {
        SmsTemplateDTO smsTemplateDTO = message.getSmsTemplateDTO();
        // 获得渠道对应的 SmsClient 客户端
        SmsClient smsClient = smsClientFactory.getSmsClient(smsTemplateDTO.getChannelId());
        Assert.notNull(smsClient, String.format("短信客户端(%d) 不存在", smsTemplateDTO.getChannelId()));
        // 发送短信
        SmsCommonResult<SmsSendRespDTO> sendResult = smsClient.sendSmsV2(message.getLogId(), message.getMobile(),
                smsTemplateDTO, message.getTemplateParams());
        smsLogService.updateSmsSendResult(message.getLogId(), sendResult.getCode(), sendResult.getMsg(),
                sendResult.getApiCode(), sendResult.getApiMsg(), sendResult.getApiRequestId(),
                sendResult.getData() != null ? sendResult.getData().getSerialNo() : null);
    }


    @Override
    public void receiveSmsStatus(String channelCode, String text) throws Throwable {
        // 获得渠道对应的 SmsClient 客户端
        SmsClient smsClient = smsClientFactory.getSmsClient(channelCode);
        Assert.notNull(smsClient, String.format("短信客户端(%s) 不存在", channelCode));
        // 解析内容
        List<SmsReceiveRespDTO> receiveResults = smsClient.parseSmsReceiveStatus(text);
        if (CollUtil.isEmpty(receiveResults)) {
            return;
        }
        // 更新短信日志的接收结果. 因为量一般不大，所以先使用 for 循环更新
        receiveResults.forEach(result -> smsLogService.updateSmsReceiveResult(result.getLogId(),
                result.getSuccess(), result.getReceiveTime(), result.getErrorCode(), result.getErrorCode()));
    }

    /**
     * 发送短信服务方法V2版本
     * 该方法负责根据提供的参数构建短信发送请求，并处理发送逻辑
     *
     * @param mobile         手机号码
     * @param userId         用户ID
     * @param userType       用户类型
     * @param nodeValue      节点值，用于确定短信内容
     * @param areaCode       区域代码，用于确定短信发送的国家或地区
     * @param isOrders       是否多订单
     * @param isTransport    是否匹配运输方式
     * @param transportId    运输方式
     * @param messageType    发送类型
     * @param templateParams 短信模板参数
     * @param smsLogId       重发短信时的原短信日志id
     * @param nodeTemplateSn 模板序号
     * @param smsNode        节点对象
     * @param receive        目的仓
     */
    @Override
    public Long sendSingleSmsV2(String mobile, Long userId, Integer userType, String nodeValue, String areaCode,
                                Integer isOrders, Integer isTransport, Integer transportId, Integer messageType, Map<String, Object> templateParams, Long smsLogId, Integer nodeTemplateSn, SmsNodeDO smsNode, String receive) {
        //首次发送
        if (smsNode == null) {
            // 创建SmsNodeDO对象并设置相关属性
            SmsNodeDO smsNodeDO = new SmsNodeDO()
                    .setNodeValue(nodeValue)
                    .setIsTransport(isTransport)
                    .setTransportId(transportId)
                    .setIsOrders(isOrders)
                    .setCountryCode(areaCode)
                    .setStatus(SystemStatusEnum.SYSTEM_STATUS_0.getValue());

            // 尝试从缓存中获取短信节点信息
            smsNode = getSmsNodeFromCache(smsNodeDO, receive);
            if (smsNode == null) {
                log.error(SMS_NODE_NOT_EXISTS.getMsg());
                return null;
            }
        }

        String smsTemplateDO = null;
        // 从缓存中获取短信模板信息
        for (int i = nodeTemplateSn; i <= 4; i++) {
            nodeTemplateSn = i;
            Long templateId = getTemplateIdBySn(smsNode, nodeTemplateSn);
            if (templateId != null) {
                String key = smsTemplateService.buildCacheKey(new SmsTemplateDO().setId(templateId).setMessageType(messageType).setStatus(SystemStatusEnum.SYSTEM_STATUS_0.getValue()));
                String smsTemplate = redisHelper.get(key);
                if (smsTemplate != null) {
                    smsTemplateDO = smsTemplate;
                    break;
                }
            }
        }

        if (smsTemplateDO == null) {
            log.error(SMS_SEND_TEMPLATE_NOT_EXISTS.getMsg());
            return null;
        }

        // 解析缓存获取短信模板信息
        SmsTemplateDO smsTemplate = JSON.parseObject(smsTemplateDO, SmsTemplateDO.class);

        // 格式化短信模板内容
        String content = smsTemplateService.formatSmsTemplateContent(smsTemplate.getContent(), templateParams);

        // 创建短信发送日志
        Long sendLogId = smsLogService.createSmsLogV2(mobile, userId, userType, true, smsTemplate,
                content, templateParams, smsTemplate.getApiTemplateId(), smsLogId, smsNode.getId(), nodeTemplateSn, messageType);

        // 构建有序的模板参数。为什么放在这个位置，是提前保证模板参数的正确性，而不是到了插入发送日志
        List<KeyValue<String, Object>> newTemplateParams = this.buildTemplateParams(smsTemplate, templateParams);

        // 创建SmsSendMessageV2对象并设置相关属性
        SmsSendMessageV2 smsSendMessage = new SmsSendMessageV2()
                .setLogId(sendLogId)
                .setMobile(mobile)
                .setSmsTemplateDTO(SmsTemplateConvert.INSTANCE.toDto(smsTemplate))
                .setTemplateParams(newTemplateParams);
        // 判断渠道
        SmsChannelDO smsChannel = smsChannelService.getSmsChannel(smsTemplate.getChannelId());
        if (smsChannel == null) {
            throw exception(SMS_CHANNEL_NOT_EXISTS);
        }
        //根据debug配置决定是否发送短信
        if ((Objects.isNull(businessProperties) || !businessProperties.isDebug()) && smsChannel.getStatus() == 0) {
            smsProducer.sendSmsSendMessageV2(smsSendMessage);
        }
        return sendLogId;
    }

    /**
     * 获取缓存节点
     */
    private SmsNodeDO getSmsNodeFromCache(SmsNodeDO smsNodeDO, String receive) {
        // 构建缓存键，用于查询短信节点缓存
        String cacheKey = smsNodeService.buildCacheKey(smsNodeDO, receive);
        String smsNodeCache = redisHelper.get(cacheKey);

        // 第一次找不到，则在全部目的仓里面找
        if (smsNodeCache == null) {
            cacheKey = smsNodeService.buildCacheKey(smsNodeDO, Constants.RECEIVE_ADDR_ALL_KEY);
            smsNodeCache = redisHelper.get(cacheKey);
        }

        // 第二次找不到，则在全部国家里面找
        if (smsNodeCache == null) {
            smsNodeDO.setCountryCode(SmsCountryCodeEnum.SMS_COUNTRY_CODE_0.getValue());
            cacheKey = smsNodeService.buildCacheKey(smsNodeDO, receive);
            smsNodeCache = redisHelper.get(cacheKey);
        }

        // 第三次找不到，则在全部国家、全部目的仓里面找
        if (smsNodeCache == null) {
            cacheKey = smsNodeService.buildCacheKey(smsNodeDO, Constants.RECEIVE_ADDR_ALL_KEY);
            smsNodeCache = redisHelper.get(cacheKey);
        }

        // 第四次找不到，则在其他国家里面找
        if (smsNodeCache == null) {
            smsNodeDO.setCountryCode(SmsCountryCodeEnum.SMS_COUNTRY_CODE_1.getValue());
            cacheKey = smsNodeService.buildCacheKey(smsNodeDO, receive);
            smsNodeCache = redisHelper.get(cacheKey);
        }

        // 第五次次找不到，则在其他国家、全部目的仓里面找
        if (smsNodeCache == null) {
            cacheKey = smsNodeService.buildCacheKey(smsNodeDO, Constants.RECEIVE_ADDR_ALL_KEY);
            smsNodeCache = redisHelper.get(cacheKey);
        }

        return smsNodeCache != null ? JSON.parseObject(smsNodeCache, SmsNodeDO.class) : null;
    }

    /**
     * 根据节点模板序号获取模板ID
     *
     * @param smsNode 短信节点信息
     * @param sn      模板序号
     * @return 对应的模板ID，如果找不到则返回null
     */
    private Long getTemplateIdBySn(SmsNodeDO smsNode, int sn) {
        switch (sn) {
            case 1:
                return smsNode.getTemplateIdOne();
            case 2:
                return smsNode.getTemplateIdTwo();
            case 3:
                return smsNode.getTemplateIdThree();
            case 4:
                return smsNode.getTemplateIdFour();
            default:
                return null;
        }
    }
}
